
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <algorithm>
#include <cstdlib>			// exit
#include <list>
#include <string>

#include <SDL/SDL.h>
#include "fastevents.h"
#include <AL/alut.h>

#include "Engine.hh"
#include "Event.hh"
#include "Exception.hh"
#include "Log.hh"
#include "Math.hh"
#include "Random.hh"
#include "Settings.hh"
#include "Timer.hh"


namespace Mf {


class Engine::Impl
{
public:

	Impl(Engine& engine) :
		mInterface(engine),
		mTimestep(0.01),
		mPrintFps(false)
	{
		// first, initialize the libraries

#if defined(_WIN32) || defined(__WIN32__)
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) != 0)
#else
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD) != 0)
#endif
		{
			const char* error = SDL_GetError();
			throw Exception(ErrorCode::SDL_INIT, error);
		}
		if (FE_Init() != 0)
		{
			const char* error = FE_GetError();
			throw Exception(ErrorCode::FASTEVENTS_INIT, error);
		}
		int argc = 1;
		char name[] = "hello";
		alutInit(&argc, (char**)&name);
		
		// now load the settings the engine needs

		Settings& settings = Settings::getInstance();

		unsigned randomSeed;
		if (settings.get("rngseed", randomSeed)) setSeed(randomSeed);
		else setSeed();

		Scalar timestep = 80.0;
		settings.get("timestep", timestep);
		mTimestep = 1.0 / timestep;

		Scalar maxFps = 40.0;
		settings.get("maxfps", maxFps);
		mMaxFps = 1.0 / maxFps;
		capFps();

		settings.get("printfps", mPrintFps);
	}

	~Impl()
	{
		// the video object must be destroyed before we can shutdown SDL
		mVideo.reset();

		alutExit();
		FE_Quit();
		SDL_Quit();
	}


	/**
	 * The main loop.  This just calls dispatchEvents(), update(), and draw()
	 * over and over again.  The timing of the update and draw are decoupled.
	 * The actual frame rate is also calculated here.  This function will return
	 * the exit code used to stop the loop.
	 */

	void run()
	{
		Scalar ticksNow = Timer::getTicks();

		Scalar nextStep = ticksNow;
		Scalar nextDraw = ticksNow;
		Scalar nextFpsUpdate = ticksNow + 1.0;

		Scalar totalTime = 0.0;
		Scalar deltaTime = 0.0;
		Scalar accumulator = mTimestep;

		mFps = 0;
		int frameAccum = 0;

		do
		{
			Scalar newTicks = Timer::getTicks();
			deltaTime = newTicks - ticksNow;
			ticksNow = newTicks;

			// don't slow the animation until 4Hz, which is unplayable anyway
			if (deltaTime >= 0.25) deltaTime = 0.25;
			accumulator += deltaTime;

			Timer::fireIfExpired(ticksNow);
			dispatchEvents();

			while (accumulator >= mTimestep)
			{
				update(totalTime, mTimestep);

				totalTime += mTimestep;
				accumulator -= mTimestep;

				nextStep += mTimestep;
			}
			if (ticksNow >= nextStep)
			{
				nextStep = ticksNow + mTimestep;
			}

			if (ticksNow >= nextDraw)
			{
				frameAccum++;

				if (ticksNow >= nextFpsUpdate) // determine the actual fps
				{
					mFps = frameAccum;
					frameAccum = 0;

					nextFpsUpdate += 1.0;
					if (ticksNow >= nextFpsUpdate)
					{
						nextFpsUpdate = ticksNow + 1.0;
					}

					if (mPrintFps)
					{
						logInfo("%d fps", mFps);
					}
				}

				draw(accumulator / mTimestep);
				mVideo->swap();

				nextDraw += mMaxFps;
				if (ticksNow >= nextDraw)
				{
					// we missed some scheduled draws, so reset the schedule
					nextDraw = ticksNow + mMaxFps;
				}
			}

			// be a good citizen and give back what you don't need
			Timer::sleep(std::min(std::min(nextStep, nextDraw),
						Timer::getNextFire()), Timer::ACTUAL);
		}
		while (!mStack.empty());
	}

	void dispatchEvents()
	{
		SDL_Event event;

		while (FE_PollEvent(&event) == 1)
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE &&
							(SDL_GetModState() & KMOD_CTRL) )
					{
						// emergency escape
						logWarning("escape forced");
						exit(1);
					}
					break;

				case SDL_VIDEORESIZE:
					mVideo->resize(event.resize.w, event.resize.h);
					break;
			}

			handleEvent(event);
		}
	}


	void update(Scalar t, Scalar dt)
	{
		for (mStackIt = mStack.begin(); mStackIt != mStack.end(); ++mStackIt)
		{
			(*mStackIt)->update(mInterface, t, dt);
		}
	}

	void draw(Scalar alpha)
	{
		// FIXME - this will crash if the layer being drawn pops itself
		std::list<LayerP>::reverse_iterator it;
		for (it = mStack.rbegin(); it != mStack.rend(); ++it)
		{
			(*it)->draw(mInterface, alpha);
		}
	}

	void handleEvent(const Event& event)
	{
		for (mStackIt = mStack.begin(); mStackIt != mStack.end(); ++mStackIt)
		{
			if ((*mStackIt)->handleEvent(mInterface, event)) break;
		}
	}


	void push(LayerP layer)
	{
		ASSERT(layer && "cannot push null layer");
		mStack.push_front(layer);
		logDebug("stack: %d [pushed %X]", mStack.size(), layer.get());
		layer->pushed(mInterface);
	}

	LayerP pop()
	{
		bool fixIt = false;
		if (mStack.begin() == mStackIt) fixIt = true;

		LayerP layer = mStack.front();
		mStack.pop_front();
		logDebug("stack: %d [popped %X]", mStack.size(), layer.get());
		layer->popped(mInterface);

		if (fixIt) mStackIt = --mStack.begin();

		return layer;
	}

	LayerP pop(Layer* layer)
	{
		bool fixIt = false;

		std::list<LayerP> layers;

		std::list<LayerP>::iterator it;
		for (it = mStack.begin(); it != mStack.end(); ++it)
		{
			layers.push_back(*it);

			if (it == mStackIt) fixIt = true;

			if ((*it).get() == layer)
			{
				++it;
				mStack.erase(mStack.begin(), it);

				for (it = layers.begin(); it != layers.end(); ++it)
				{
					(*it)->popped(mInterface);
					logDebug("stack: %d [popped %X]", mStack.size(), (*it).get());
				}

				if (fixIt) mStackIt = --mStack.begin();

				return layers.back();
			}
		}

		return LayerP();
	}

	void clear()
	{
		mStack.clear();
		mStackIt = mStack.begin();
		logDebug("stack: 0 [cleared]");
	}


	void capFps()
	{
		if (mMaxFps < mTimestep)
		{
			logWarning("capping maximum fps to timestep (%f)", mTimestep);
			mMaxFps = mTimestep;
		}
	}


	Engine&						mInterface;
	VideoP						mVideo;
	Dispatch					mDispatch;

	std::list<LayerP>			mStack;
	std::list<LayerP>::iterator	mStackIt;

	Scalar						mTimestep;
	Scalar						mMaxFps;

	int							mFps;
	bool						mPrintFps;
};


Engine::Engine() :
	// pass through
	mImpl(new Engine::Impl(*this)) {}

Engine& Engine::getInstance()
{
	static Engine engine;
	return engine;
}


void Engine::setVideo(VideoP video)
{
	// pass through
	mImpl->mVideo = video;
}

VideoP Engine::getVideo() const
{
	return mImpl->mVideo;
}


void Engine::setTimestep(int ts)
{
	mImpl->mTimestep = 1.0 / Scalar(ts);
	mImpl->capFps();
}

int Engine::getTimestep() const
{
	return int(1.0 / mImpl->mTimestep);
}


void Engine::setMaxFps(int maxFps)
{
	mImpl->mMaxFps = 1.0 / Scalar(maxFps);
	mImpl->capFps();
}

int Engine::getMaxFps() const
{
	return int(1.0 / mImpl->mMaxFps);
}


int Engine::getFps() const
{
	return mImpl->mFps;
}


void Engine::push(LayerP layer)
{
	// pass through
	mImpl->push(layer);
}

LayerP Engine::pop()
{
	// pass through
	return mImpl->pop();
}

LayerP Engine::pop(Layer* layer)
{
	// pass through
	return mImpl->pop(layer);
}

void Engine::clear()
{
	// pass through
	mImpl->clear();
}

int Engine::getSize() const
{
	return mImpl->mStack.size();
}


void Engine::run()
{
	// pass through
	return mImpl->run();
}


Dispatch::Handler Engine::addHandler(const std::string& event,
		const Dispatch::Function& callback)
{
	return mImpl->mDispatch.addHandler(event, callback);
}

Dispatch::Handler Engine::addHandler(const std::string& event,
		const Dispatch::Function& callback, Dispatch::Handler handler)
{
	return mImpl->mDispatch.addHandler(event, callback, handler);
}

void Engine::dispatch(const std::string& event,
		const Dispatch::Message* message)
{
	mImpl->mDispatch.dispatch(event, message);
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

